from django.db import models
from audit_log.models.fields import LastUserField
from audit_log.models.managers import AuditLog
import datetime

class ProductCategory(models.Model):
    name = models.CharField(max_length=150, primary_key = True)
    description = models.TextField()
    
    audit_log = AuditLog()
    
    def __unicode__(self):
        return self.name
    
class Product(models.Model):
    name = models.CharField(max_length = 150)
    description = models.TextField()
    price = models.DecimalField(max_digits = 10, decimal_places = 2)
    category = models.ForeignKey(ProductCategory)
    
    audit_log = AuditLog()
    
    
    def __unicode__(self):
        return self.name

class WarehouseEntry(models.Model):
    product = models.ForeignKey(Product)
    quantity = models.DecimalField(max_digits = 10, decimal_places = 2)
    
    audit_log = AuditLog()
    
    class Meta:
        app_label = "warehouse"


class SaleInvoice(models.Model):
    date = models.DateTimeField(default = datetime.datetime.now)

    audit_log = AuditLog(exclude = ['date',])
    
    
    def __unicode__(self):
        return str(self.date)

class SoldQuantity(models.Model):
    product = models.ForeignKey(Product)
    quantity = models.DecimalField(max_digits = 10, decimal_places = 2)
    sale = models.ForeignKey(SaleInvoice)

    audit_log = AuditLog()
    
    
    def __unicode__(self):
        return "%s X %s"%(self.product.name, self.quantity)


class Widget(models.Model):
    name = models.CharField(max_length = 100)

class ExtremeWidget(Widget):
    special_power = models.CharField(max_length = 100)
    
    audit_log = AuditLog()


from django.test import TestCase
from testproject.store.models import Product, WarehouseEntry, ProductCategory, ExtremeWidget
from django.test.client import Client

def _setup_admin():
    from django.contrib.auth.models import User
    User.objects.all().delete()
    admin = User(username = "admin", is_staff = True, is_superuser = True)
    admin.set_password("admin")
    admin.save()
    

class EntryManagerSelectTest(TestCase):
    fixtures = ['test_data.json']
    
    def test_model_level_select(self):
        self.failUnlessEqual(Product.audit_log.all().count(), 10)
        
    def test_instance_level_select(self):
        self.failUnlessEqual(Product.objects.get(pk = 4).audit_log.all().count(), 3)
        self.failUnlessEqual(Product.objects.get(pk = 1).audit_log.all().count(), 2)
        

class LogEntryMetaOptionsTest(TestCase):
    
    def test_app_label(self):
        self.failUnlessEqual(Product.audit_log.model._meta.app_label, Product._meta.app_label)
        self.failUnlessEqual(WarehouseEntry.audit_log.model._meta.app_label, WarehouseEntry._meta.app_label)
    
    def test_table_name(self):
        self.failUnlessEqual(Product.audit_log.model._meta.db_table, "%sauditlogentry"%Product._meta.db_table)
        self.failUnlessEqual(WarehouseEntry.audit_log.model._meta.db_table, "%sauditlogentry"%WarehouseEntry._meta.db_table)


class LoggingTest(TestCase):
    
    def setup_client(self):
        c = Client()
        c.login(username = "admin", password = "admin")
        return c
    
    def test_logging_insert_update(self):
        _setup_admin()
        c = self.setup_client()
        c.post('/admin/store/productcategory/add/', {'name': 'Test Category', 'description': 'Test description'})
        self.failUnlessEqual(ProductCategory.objects.all().count(), 1)
        category = ProductCategory.objects.all()[0]
        self.failUnlessEqual(category.audit_log.all()[0].name, category.name)
        self.failUnlessEqual(category.audit_log.all()[0].description, category.description)
        self.failUnlessEqual(category.audit_log.all()[0].action_type, "I")
        self.failUnlessEqual(category.audit_log.all()[0].action_user.username, "admin")
    
        c.post('/admin/store/productcategory/%s/'%'Test Category', {'name': 'Test Category new name', 'description': 'Test description'})
        category = ProductCategory.objects.get(pk = "Test Category new name")
        self.failUnlessEqual(category.audit_log.all().count(), 1)
        self.failUnlessEqual(category.audit_log.all()[0].name, "Test Category new name")
        self.failUnlessEqual(category.audit_log.all()[0].action_type, "I")
        
        c.post('/admin/store/productcategory/%s/'%'Test Category new name', {'name': 'Test Category new name', 
                                                                            'description': 'Test modified description'})
        category = ProductCategory.objects.get(pk = "Test Category new name")
        self.failUnlessEqual(category.audit_log.all().count(), 2)
        self.failUnlessEqual(category.audit_log.all()[0].description, "Test modified description")
        self.failUnlessEqual(category.audit_log.all()[0].action_type, "U")
    
    def test_logging_delete(self):
        _setup_admin()
        c = self.setup_client()
        c.post('/admin/store/productcategory/add/', {'name': 'Test', 'description': 'Test description'})
        self.failUnlessEqual(ProductCategory.objects.all().count(), 1)
        c.post('/admin/store/productcategory/Test/delete/', {'post': 'yes'})
        self.failUnlessEqual(ProductCategory.objects.all().count(), 0)
        self.failUnlessEqual(ProductCategory.audit_log.all().count(), 2)
        self.failUnlessEqual(ProductCategory.audit_log.all()[0].action_type, "D")        
        
        
    def test_logging_inherited(self):
        _setup_admin()
        c = Client()
        c.login(username = "admin", password = "admin")
        c.post('/admin/store/extremewidget/add/', {'name': 'Test name', 'special_power': 'Testpower'})
        widget = ExtremeWidget.objects.all()[0]
        self.failUnlessEqual(widget.audit_log.all()[0].name, 'Test name')
        self.failUnlessEqual(hasattr(widget.audit_log.all()[0], 'special_power'), True)
        self.failUnlessEqual(widget.audit_log.all()[0].special_power, "Testpower")
        
        


import copy
import datetime
from django.db import models
from django.utils.functional import curry
from django.utils.translation import ugettext_lazy as _



from audit_log.models.fields import LastUserField

class LogEntryObjectDescriptor(object):
    def __init__(self, model):
        self.model = model
    
    def __get__(self, instance, owner):
        kwargs = dict((f.attname, getattr(instance, f.attname))
                    for f in self.model._meta.fields
                    if hasattr(instance, f.attname))
        return self.model(**kwargs)
        
      
class AuditLogManager(models.Manager):
    def __init__(self, model, instance = None):
        super(AuditLogManager, self).__init__()
        self.model = model
        self.instance = instance
    
    def get_query_set(self):
        if self.instance is None:
            return super(AuditLogManager, self).get_query_set()
        
        f = {self.instance._meta.pk.name : self.instance.pk}
        return super(AuditLogManager, self).get_query_set().filter(**f)
    
            
class AuditLogDescriptor(object):
    def __init__(self, model, manager_class):
        self.model = model
        self._manager_class = manager_class
    
    def __get__(self, instance, owner):
        if instance is None:
            return self._manager_class(self.model)
        return self._manager_class(self.model, instance)

class AuditLog(object):
    
    manager_class = AuditLogManager
    
    def __init__(self, exclude = []):
        self._exclude = exclude
    
    def contribute_to_class(self, cls, name):
        self.manager_name = name
        models.signals.class_prepared.connect(self.finalize, sender = cls)
    
    
    def create_log_entry(self, instance, action_type):
        manager = getattr(instance, self.manager_name)
        attrs = {}
        for field in instance._meta.fields:
            if field.attname not in self._exclude:
                attrs[field.attname] = getattr(instance, field.attname)
        manager.create(action_type = action_type, **attrs)
    
    def post_save(self, instance, created, **kwargs):
        self.create_log_entry(instance, created and 'I' or 'U')
    
    
    def post_delete(self, instance, **kwargs):
        self.create_log_entry(instance,  'D')
    
    
    def finalize(self, sender, **kwargs):
        log_entry_model = self.create_log_entry_model(sender)
        
        models.signals.post_save.connect(self.post_save, sender = sender, weak = False)
        models.signals.post_delete.connect(self.post_delete, sender = sender, weak = False)
        
        descriptor = AuditLogDescriptor(log_entry_model, self.manager_class)
        setattr(sender, self.manager_name, descriptor)
    
    def copy_fields(self, model):
        """
        Creates copies of the fields we are keeping
        track of for the provided model, returning a 
        dictionary mapping field name to a copied field object.
        """
        fields = {'__module__' : model.__module__}
        
        for field in model._meta.fields:
            
            if not field.name in self._exclude:
                
                field  = copy.deepcopy(field)
            
                if isinstance(field, models.AutoField):
                    #we replace the AutoField of the original model
                    #with an IntegerField because a model can
                    #have only one autofield.
                
                    field.__class__ = models.IntegerField
                
                if field.primary_key:
                    field.serialize = True
            
                if field.primary_key or field.unique:
                    #unique fields of the original model
                    #can not be guaranteed to be unique
                    #in the audit log entry but they
                    #should still be indexed for faster lookups.
                
                    field.primary_key = False
                    field._unique = False
                    field.db_index = True
                    
                
                if field.rel and field.rel.related_name:
                    field.rel.related_name = '_auditlog_%s' % field.rel.related_name
            

                
                fields[field.name] = field
            
        return fields
    

    
    def get_logging_fields(self, model):
        """
        Returns a dictionary mapping of the fields that are used for
        keeping the acutal audit log entries.
        """
        rel_name = '_%s_audit_log_entry'%model._meta.object_name.lower()
        
        def entry_instance_to_unicode(log_entry):
            try:
                result = u'%s: %s %s at %s'%(model._meta.object_name, 
                                                log_entry.object_state, 
                                                log_entry.get_action_type_display().lower(),
                                                log_entry.action_date,
                                                
                                                )
            except AttributeError:
                result = u'%s %s at %s'%(model._meta.object_name,
                                                log_entry.get_action_type_display().lower(),
                                                log_entry.action_date
                                                
                                                )
            return result
        
        return {
            'action_id' : models.AutoField(primary_key = True),
            'action_date' : models.DateTimeField(default = datetime.datetime.now),
            'action_user' : LastUserField(related_name = rel_name),
            'action_type' : models.CharField(max_length = 1, choices = (
                ('I', _('Created')),
                ('U', _('Changed')),
                ('D', _('Deleted')),
            )),
            'object_state' : LogEntryObjectDescriptor(model),
            '__unicode__' : entry_instance_to_unicode,
        }
            
    
    def get_meta_options(self, model):
        """
        Returns a dictionary of fileds that will be added to
        the Meta inner class of the log entry model.
        """
        return {
            'ordering' : ('-action_date',),
            'app_label' : model._meta.app_label,
        }
    
    def create_log_entry_model(self, model):
        """
        Creates a log entry model that will be associated with
        the model provided.
        """
        
        attrs = self.copy_fields(model)
        attrs.update(self.get_logging_fields(model))
        attrs.update(Meta = type('Meta', (), self.get_meta_options(model)))
        name = '%sAuditLogEntry'%model._meta.object_name
        return type(name, (models.Model,), attrs)
        


class FieldRegistry(object):
    _registry = {}
    
    def __init__(self, fieldcls):
        self._fieldcls = fieldcls

    
    def add_field(self, model, field):
        reg = self.__class__._registry.setdefault(self._fieldcls, {}).setdefault(model, [])
        reg.append(field)
    
    def get_fields(self, model):
        return self.__class__._registry.setdefault(self._fieldcls, {}).get(model, [])
    
    def __contains__(self, model):
        return model in self.__class__._registry.setdefault(self._fieldcls, {})


from django.contrib import admin
from models import *

class CategoryAdmin(admin.ModelAdmin):
    pass


class ProductAdmin(admin.ModelAdmin):
    pass


class QuantityInline(admin.TabularInline):
    model = SoldQuantity
    


class InvoiceAdmin(admin.ModelAdmin):
    list_display = ('date',)    
    inlines = [
        QuantityInline,
    ]
    
class ExtremeWidgetAdmin(admin.ModelAdmin):
    list_display  = ('name', 'special_power')
    

admin.site.register(ProductCategory, CategoryAdmin)
admin.site.register(Product, ProductAdmin)
admin.site.register(WarehouseEntry, admin.ModelAdmin)
admin.site.register(SaleInvoice, InvoiceAdmin)
admin.site.register(ExtremeWidget, ExtremeWidgetAdmin)



from django.db import models
from django.contrib.auth.models import User
from audit_log import registration

class LastUserField(models.ForeignKey):
    """
    A field that keeps the last user that saved an instance
    of a model. None will be the value for AnonymousUser.
    """
    
    def __init__(self, to = User, null = True,  **kwargs):
        super(LastUserField, self).__init__(to = to, null = null, **kwargs)
    
    def contribute_to_class(self, cls, name):
        super(LastUserField, self).contribute_to_class(cls, name)
        registry = registration.FieldRegistry(self.__class__)
        registry.add_field(cls, self)


from django.db.models import signals
from django.utils.functional import curry

from audit_log import registration
from audit_log.models import fields

class UserLoggingMiddleware(object):
    def process_request(self, request):
        if not request.method in ('GET', 'HEAD', 'OPTIONS', 'TRACE'):
            if hasattr(request, 'user') and request.user.is_authenticated():
                user = request.user
            else:
                user = None
        
            update_users = curry(self.update_users, user)
            signals.pre_save.connect(update_users,  dispatch_uid = (self.__class__, request,), weak = False)
    
    def process_response(self, request, response):
        signals.pre_save.disconnect(dispatch_uid =  (self.__class__, request,))
        return response
    
    def update_users(self, user, sender, instance, **kwargs):
        registry = registration.FieldRegistry(fields.LastUserField)
        if sender in registry:
            for field in registry.get_fields(sender):
                setattr(instance, field.name, user)
    



import os, sys
from setuptools import setup, find_packages

import audit_log
from audit_log import VERSION, __version__

if VERSION[-1] == 'final':
    STATUS = ['Development Status :: 5 - Production/Stable']
elif 'beta' in VERSION[-1]:
    STATUS = ['Development Status :: 4 - Beta']
else:
    STATUS = ['Development Status :: 3 - Alpha']

def get_readme():
    try:
        return  open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
    except IOError:
        return ''

setup(
    name = 'django-audit-log',
    version = __version__,
    packages = find_packages(exclude = ['testproject']),
    author = 'Vasil Vangelovski',
    author_email = 'vvangelovski@gmail.com',
    license = 'New BSD License (http://www.opensource.org/licenses/bsd-license.php)',
    description = 'Audit trail for django models',
    long_description = get_readme(),
    url = 'https://github.com/Atomidata/django-audit-log',
    download_url = 'https://github.com/Atomidata/django-audit-log/downloads',
    include_package_data = True,
    zip_safe = False,
    
    classifiers = STATUS + [
       'Environment :: Plugins',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
